/*************************************************************************/
/*************************************************************************/
/* H. Timucin OZDEMIR                                                    */
/* Date : July, 1997                                                     */
/* File Name : RotationFinder.cc                                         */
/*                                                                       */
/* Updated at : July, 1997                                               */
/*************************************************************************/
/*  Finds the Rotations Randomly
*/
/*************************************************************************/

#include <stdlib.h>

#include "FlightGraph.h"
#include "City.h"
#include "AirScheduleProblem.h"
#include "RotationFinder.h"
#include "Stack.h"


/*
//
*/

RotationFinder::RotationFinder(FlightGraph _flightGraph, City *baseCities,
		     unsigned _bSize, AirScheduleProblem * _myProblem)
{

   myFlightGraph = _flightGraph;

   myBaseCities = new City[_bSize];
   NumberOfBaseCities = _bSize;

   for(int i=0;i<NumberOfBaseCities;i++) 
       myBaseCities[i] = baseCities[i];

   myProblem = _myProblem;
   DEBUG = 0;

   if(DEBUG) cout << " RotationFinder is defined"<< endl;

}// end of constructor

/*
//
*/

RotationFinder::~RotationFinder() 
{
   if(DEBUG) cout << " RotationFinder is deallocated"<< endl;
}

/*
//
*/

List<ListOfFlights> RotationFinder::findRotations() 
{
  FlightGraph tmpFlightGraph = myFlightGraph;
  City *tmpBaseCities = new City[NumberOfBaseCities];
  int i;
  unsigned tmpNumberOfBaseCities = NumberOfBaseCities;

  /********************/
  /** Flight Schedule */
  /********************/
  List<ListOfFlights> schedule;

  /**************************/
  /** Is Schedule Complete? */
  /**************************/
  int scheduleDone = 0;

  /*********************/
  /** Is it printable? */
  /*********************/
  int printTime = 0;

  unsigned indexBaseCity;
  unsigned indexFlight;
  unsigned start;

  List<Flight> currentFList;
  Flight flight;

  mystack<Flight> flightStack(80);

  /***************************************************/
  /** Copy BaseCities to temporary Base Cities array */
  /***************************************************/
  for(i=0;i<NumberOfBaseCities;i++) 
     tmpBaseCities[i]=myBaseCities[i];


  if(DEBUG) cout << " Before while(!scheduleDone) " << endl;
  while(!scheduleDone && tmpFlightGraph.getNumberOfFlights()>0)
  {
    /********************************/
    /** pick the base city randomly */
    /********************************/
    indexBaseCity = rand() % tmpNumberOfBaseCities;

    indexBaseCity = tmpBaseCities[indexBaseCity].getIndex();

    if(DEBUG) cout << " indexBaseCity is " << indexBaseCity << endl;

    /*********************************************************/
    /** Get the List of Flight taking off from indexBaseCity */
    /*********************************************************/
    currentFList = tmpFlightGraph.getFlightList(indexBaseCity); 
    if(DEBUG) cout << " currentFList is " << endl; 

    /**********************************************************/
    /** pick the flight randomly from the current flight list */
    /**********************************************************/
    indexFlight = rand() % currentFList.getSize();
    if(DEBUG) cout << " indexFlight is " << indexFlight << endl; 

    /************************************************/
    /** Get this flight's information from the list */
    /************************************************/
    flight = currentFList.getElement(indexFlight);

    /*********************************************************/
    /** Get the index of the city where this flight took off */
    /*********************************************************/
    start = flight.getFromIndex();

    printTime = 0;

    if(DEBUG) cout << " Before while(!printTime) " << endl;
    while(!printTime)
    {
       /*****************************************/
       /** push the current flight to the stack */
       /*****************************************/

       flightStack.push(flight);

       //if(DEBUG) cout << " after push " << endl;

       /***************************************/
       /** Gets the index of destination city */
       /***************************************/

       unsigned current = flight.getToIndex();

       unsigned k = 0;

       /*********************************************/
       /** Gets the list of flights taking off from */
       /** the destination city                     */
       /*********************************************/

       List<Flight> tmpFList = tmpFlightGraph.getFlightList(current);

       List<Flight> tmpPickList;

       /******************************************/
       /** Pick the possible flights that can be */
       /** assigned from this city               */
       /******************************************/

       for(int j=0;j<tmpFList.getSize();j++)
         if( flight.getArrivalDate() < 
	     tmpFList.getElement(j).getDepartureDate() )
         {
	    tmpPickList.addElement(tmpFList.getElement(j));
	    k++; 
         }// end of if

       //if(DEBUG) cout << " tmpPickList size " << tmpPickList.getSize() << endl;

       /*************************************************/
       /** If there are some flights that we can assign */
       /*************************************************/

       if(tmpPickList.getSize()>0)
       {  
          /******************************************/
	  /** Pick randomly one of possible flights */
          /******************************************/
	  unsigned pickIndex = rand() % tmpPickList.getSize();

          /***************************************/
	  /** Get the new flight on the rotation */
          /***************************************/
	  flight = tmpPickList.getElement(pickIndex);
       } 
       else printTime = 1; // we could not find a connection flight

    }// end of while print is not POSSIBLE

    //if(DEBUG) 
    //   cout << " Now it is printTime size " << endl;

    /****************************************************/
    /** Now we have a some sort of rotation composed of */
    /** several flights hopefully.                      */
    /****************************************************/

    List<Flight> rotation;

    //if(DEBUG) 
    //   cout << " rotation is defined " << endl;

    Flight *ptr  = flightStack.pop();

    if(DEBUG) 
    {  cout << " Popped flight is " << *ptr << endl;
       cout << " Size of Stack is " << flightStack.size_of()+1<< endl;
    }// end of if

    while( ptr != NULL)
    {
       if(DEBUG) cout << " Before remove from FlightGraph " << endl;

       /********************************************/
       /** remove this flight from the FlightGraph */
       /********************************************/
       tmpFlightGraph.removeFlight(*ptr);
       if(DEBUG) cout << " After remove from FlightGraph " << endl;

       /************************************/
       /** Add this flight to the rotation */
       /************************************/
       rotation.addElement(*ptr);
       if(DEBUG) cout << " After add rotation " << endl;

       /*************************/
       /** Pop the nest element */
       /*************************/
       ptr = flightStack.pop();

       //if(DEBUG) 
       //  if(ptr!=NULL) cout << " Popped flight is " << *ptr << endl;
       //  else  cout << " Popped flight is NULL "<< endl;

    }// end of while

    if(DEBUG) 
    {  cout << " rotation size " << rotation.getSize() << endl;
       rotation.printMe(cout);
       cout << endl;
    }

    /**********************************************/
    /** If there are some flights in the rotation */
    /** then add this rotation to the schedule.   */ 
    /**********************************************/
    if(rotation.getSize()>0) schedule.addElement(rotation);

    currentFList = tmpFlightGraph.getFlightList(indexBaseCity);
    if(DEBUG && currentFList.getSize() == 0 )
       cout << " City " << indexBaseCity << " has 0 elements " << endl;

    if( rotation.getSize() == 0 || currentFList.getSize() == 0 )
    {  
        /**********************************************/
	/** Bad luck! We could not build a rotation.  */
	/** We need to check whether we exhausted the */
	/** base cities or not.                       */
	/** If we did, then it is time to include     */
	/** several stand by cities as a base city    */
	/** and try to solve the problem              */
        /**********************************************/
	if(DEBUG) 
	   cout << " We have plenty of flights " << tmpFlightGraph.getNumberOfFlights() << endl;

        /************************************************/
	/** If there are some flights left in the graph */
        /************************************************/
	if(tmpFlightGraph.getNumberOfFlights()>0)
	{
	  if(DEBUG)
	     cout << " We have plenty of flights " << tmpFlightGraph.getNumberOfFlights() << endl;

	  i=0;
	  while( tmpNumberOfBaseCities > 0 &&
		 i < tmpNumberOfBaseCities )
          {
            /************************************/
	    /** Gets the index of the base city */
            /************************************/
	    unsigned cityIndex = tmpBaseCities[i].getIndex();

            /*******************************************/
	    /** If there is no flights taking off from */
	    /** cityIndex'th city.                     */
            /*******************************************/
	    if( tmpFlightGraph.getFlightList(cityIndex).getSize()==0)
	    {
	       if(DEBUG) cout << " City " << i << " should be removed " << endl;
	       unsigned j=i+1;

               /*******************************************/
	       /** remove this from the base cities array */
               /*******************************************/
	       while(j<tmpNumberOfBaseCities)
	       {  tmpBaseCities[j-1] = tmpBaseCities[j]; j++; }

               /***************************************/
	       /** decrease the number of base cities */
               /***************************************/
	       tmpNumberOfBaseCities--;

	    } else i++;

	  }// end of while

          /*****************************************/
	  /** If we have no base city anymore then */
	  /** we should promote some of the cities */
	  /** to this status.                      */
          /*****************************************/
	  if(tmpNumberOfBaseCities==0)
	  {
             /************************************/
	     /** get the total number of cities. */
             /************************************/
	     unsigned numberOfCities= tmpFlightGraph.getNumberOfCities();

	     unsigned max = 0;
	     int indx = -1;

             /***********************************************/
	     /** Let's try to pick the city who has maximum */
	     /** number of flights taking off               */
             /***********************************************/
	     for(i=0;i <numberOfCities; i++)
	     {
               /******************************************/
	       /** Gets the number of flights taking off */
	       /** from city i                           */
               /******************************************/
               unsigned k = tmpFlightGraph.getFlightList(i).getSize();

               /******************************************/
	       /** If this number is bigger than the max */
	       /** then make this city max               */
               /******************************************/
	       if(k>max) { max=k;indx=i; }

	     }// end of for

	     if(indx==-1) 
		// If it enters here, sure it is weird.
		scheduleDone = 1;
	     else
	      {
                 /******************************************/
		 /** Place this city to the BaseCity Array */
                 /******************************************/
		 tmpNumberOfBaseCities = 1;

                 /******************************************/
		 /** Ask the AirScheduleProblem to get the */ 
		 /** City info for us.                     */
                 /******************************************/
		 tmpBaseCities[0] = *(myProblem->getCityWithIndex(indx));

	      }

	  }// end of if

	}// end of if

     }// end of else

     if(tmpFlightGraph.getNumberOfFlights()==0)
     {  scheduleDone = 1;
	if(DEBUG) cout << " Schedule is DONE " << endl;
     }

  }// end of while Schedule is NOT COMPLETE

  return schedule;

}// end of findRotations()



